#!/usr/bin/env python
# ct_statevector_tools.py

"""
Author : peters 

Revision History:
File created on 28 Jul 2010.

"""

import os
import sys
sys.path.append('../../')

import logging
import datetime
from da.baseclasses.statevector import EnsembleMember, StateVector
import numpy as np

identifier = 'CarbonTracker Statevector '
version    = '0.0'

################### Begin Class CtStateVector ###################

class CtStateVector(StateVector):
    """ This is a StateVector object for CarbonTracker. It has a private method to make new ensemble members """

    def getid(self):
        return identifier

    def getversion(self):
        return version

    def GetCovariance(self,DaSystem):
        """ Make a new ensemble from specified matrices, the attribute lag refers to the position in the state vector. 
            Note that lag=1 means an index of 0 in python, hence the notation lag-1 in the indexing below.
            The argument is thus referring to the lagged state vector as [1,2,3,4,5,..., nlag]
        """    

        import Nio

        # Get the needed matrices from the specified covariance files

        file_ocn_cov = DaSystem.da_settings['ocn.covariance'] 
        file_bio_cov = DaSystem.da_settings['bio.covariance'] 

        for file in [file_ocn_cov,file_bio_cov]:

            if not os.path.exists(file):

                msg = "Cannot find the specified file %s" % file ; logging.error(msg)
                raise IOError,msg
            else:

                msg = "Using covariance file: %s" % file ; logging.info(msg)

        f_ocn   = Nio.open_file(file_ocn_cov,'r')
        f_bio   = Nio.open_file(file_bio_cov,'r')

        cov_ocn = f_ocn.variables['CORMAT'].get_value()
        cov_bio = f_bio.variables['qprior'].get_value()

        dummy   = f_ocn.close()
        dummy   = f_bio.close()

        dummy   = logging.debug("Succesfully closed files after retrieving prior covariance matrices")

        # Once we have the matrices, we can start to make the full covariance matrix, and then decompose it

        fullcov = np.zeros((self.nparams,self.nparams),float)

        nocn    = cov_ocn.shape[0]
        nbio    = cov_bio.shape[0]

        fullcov[0:nbio,0:nbio]                  = cov_bio
        fullcov[nbio:nbio+nocn,nbio:nbio+nocn]  = 0.16 *np.dot(cov_ocn,cov_ocn)
        fullcov[nocn+nbio,nocn+nbio]            = 1.e-10

        # Visually inspect full covariance matrix if verbose

        try:
            plt.imshow(fullcov)
            plt.colorbar()
            plt.savefig('fullcovariancematrix.png')
            plt.close('all')
        except:
            pass

        return fullcov

################### End Class CtStateVector ###################


if __name__ == "__main__":

    import os
    import sys
    from da.tools.general import StartLogger 
    from da.tools.initexit import CycleControl 
    import numpy as np
    import da.tools.rc as rc

    opts = ['-v']
    args = {'rc':'../../da.rc','logfile':'da_initexit.log','jobrcfilename':'test.rc'}

    StartLogger()
    DaCycle = CycleControl(opts,args)

    dummy = EnsembleMember(0)
    print dummy

    DaCycle.Initialize()
    print DaCycle

    StateVector = PrepareState(DaCycle)

    members =  StateVector.EnsembleMembers[1]

    members[0].WriteToFile(DaCycle.da_settings['dir.input'])

    StateVector.Propagate()

    savedir         = DaCycle.da_settings['dir.output']
    filename        = os.path.join(savedir,'savestate.nc')

    StateVector.WriteToFile(filename)


    savedir         = DaCycle.da_settings['dir.output']
    filename        = os.path.join(savedir,'savestate.nc')

    StateVector.ReadFromFile(filename)

